@@ -27,6 +27,43 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: dict.h,v $
+ * Revision 1.34  2004/11/23 11:33:08  csoutheren
+ * Fixed problem with RemoveAt returning invalid pointer in some cases,
+ * and added extra documentation on this case.
+ * Thanks to Diego Tartara for pointing out this potential problem
+ *
+ * Revision 1.33  2004/04/09 03:42:34  csoutheren
+ * Removed all usages of "virtual inline" and "inline virtual"
+ *
+ * Revision 1.32  2004/04/03 23:53:09  csoutheren
+ * Added various changes to improce compatibility with the Sun Forte compiler
+ *   Thanks to Brian Cameron
+ * Added detection of readdir_r version
+ *
+ * Revision 1.31  2003/09/17 01:18:02  csoutheren
+ * Removed recursive include file system and removed all references
+ * to deprecated coooperative threading support
+ *
+ * Revision 1.30  2003/03/31 01:23:56  robertj
+ * Added ReadFrom functions for standard container classes such as
+ *   PIntArray and PStringList etc
+ *
+ * Revision 1.29  2002/10/04 01:47:29  robertj
+ * Added various increment and decrement operators to POrdinalKey.
+ *
+ * Revision 1.28  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.27  2002/06/14 13:22:12  robertj
+ * Fixed ability to remove elements from a PSet by value.
+ * Added by value add and remove functions to a PSet.
+ * Added a POrdinalSet class.
+ * Fixed some documentation.
+ *
+ * Revision 1.26  2002/02/06 00:53:25  robertj
+ * Fixed missing const on PSet::Contains and operator[], thanks Francisco Olarte Sanz
+ *
  * Revision 1.25  1999/11/30 00:22:54  robertj
  * Updated documentation for doc++
  *
@@ -107,7 +144,7 @@
  *
  */
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -128,8 +165,12 @@ class POrdinalKey : public PObject
     /** Create a new key for ordinal index values.
      */
     PINLINE POrdinalKey(
-      PINDEX newKey   /// Ordinal index value to use as a key.
+      PINDEX newKey = 0   /// Ordinal index value to use as a key.
     );
+
+    /**Operator to assign the ordinal.
+      */
+    PINLINE POrdinalKey & operator=(PINDEX);
   //@}
 
   /**@name Overrides from class PObject */
@@ -170,6 +211,30 @@ class POrdinalKey : public PObject
     /** Operator so that a POrdinalKey can be used as a PINDEX value.
      */
     PINLINE operator PINDEX() const;
+
+    /**Operator to pre-increment the ordinal.
+      */
+    PINLINE PINDEX operator++();
+
+    /**Operator to post-increment the ordinal.
+      */
+    PINLINE PINDEX operator++(int);
+
+    /**Operator to pre-decrement the ordinal.
+      */
+    PINLINE PINDEX operator--();
+
+    /**Operator to post-decrement the ordinal.
+      */
+    PINLINE PINDEX operator--(int);
+
+    /**Operator to add the ordinal.
+      */
+    PINLINE POrdinalKey & operator+=(PINDEX);
+
+    /**Operator to subtract from the ordinal.
+      */
+    PINLINE POrdinalKey & operator-=(PINDEX );
   //@}
 
   private:
@@ -288,6 +353,7 @@ class PHashTable : public PCollection
     // Member variables
     class Element {
       public:
+        friend class Table;
         PObject * key;
         PObject * data;
         Element * next;
@@ -404,6 +470,16 @@ class PAbstractSet : public PHashTable
       const PObject * obj   /// Existing object to remove from the collection.
     );
 
+    /**Remove an object at the specified index. If the #AllowDeleteObjects#
+       option is set then the object is also deleted.
+
+       @return
+       pointer to the object being removed, or NULL if it was deleted.
+     */
+    virtual PObject * RemoveAt(
+      PINDEX index   /// Index position in collection to place the object.
+    );
+
     /**This function is the same as PHashTable::AbstractGetKeyAt().
 
        @return
@@ -457,17 +533,6 @@ class PAbstractSet : public PHashTable
       const PObject & obj   /// Object to find equal value.
     ) const;
   //@}
-
-
-  private:
-    virtual PObject * RemoveAt(
-      PINDEX index   // Index position in collection to place the object.
-    );
-    /* This function is meaningless and will assert if executed.
-
-       @return
-       Always NULL.
-     */
 };
 
 
@@ -512,7 +577,7 @@ template <class T> class PSet : public PAbstractSet
 
   /**@name New functions for class */
   //@{
-    /**Include the spcified object into the set. If the objects value is
+    /**Include the specified object into the set. If the objects value is
        already in the set then the object is {\bf not} included. If the
        AllowDeleteObjects option is set then the #obj# parameter is
        also deleted.
@@ -525,6 +590,17 @@ template <class T> class PSet : public PAbstractSet
       const T * obj   // New object to include in the set.
     ) { Append((PObject *)obj); }
 
+    /**Include the specified objects value into the set. If the objects value
+       is already in the set then the object is {\bf not} included.
+
+       The object values are compared, not the pointers.  So the objects in
+       the collection must correctly implement the #PObject::Compare()#
+       function. The hash table is used to locate the entry.
+     */
+    PSet & operator+=(
+      const T & obj   // New object to include in the set.
+    ) { Append(obj.Clone()); return *this; }
+
     /**Remove the object from the set. If the AllowDeleteObjects option is set
        then the object is also deleted.
 
@@ -536,17 +612,16 @@ template <class T> class PSet : public PAbstractSet
       const T * obj   // New object to exclude in the set.
     ) { Remove(obj); }
 
-    /**Determine if the value of the object is contained in the set. The
-       object values are compared, not the pointers.  So the objects in the
-       collection must correctly implement the #PObject::Compare()#
-       function. The hash table is used to locate the entry.
+    /**Remove the objects value from the set. If the AllowDeleteObjects
+       option is set then the object is also deleted.
 
-       @return
-       TRUE if the object value is in the set.
+       The object values are compared, not the pointers.  So the objects in
+       the collection must correctly implement the #PObject::Compare()#
+       function. The hash table is used to locate the entry.
      */
-    BOOL operator[](
-      const T & key  /// Key to look for in the set.
-    ) { return AbstractContains(key); }
+    PSet & operator-=(
+      const T & obj   // New object to exclude in the set.
+    ) { RemoveAt(GetValuesIndex(obj)); return *this; }
 
     /**Determine if the value of the object is contained in the set. The
        object values are compared, not the pointers.  So the objects in the
@@ -558,7 +633,19 @@ template <class T> class PSet : public PAbstractSet
      */
     BOOL Contains(
       const T & key  /// Key to look for in the set.
-    ) { return AbstractContains(key); }
+    ) const { return AbstractContains(key); }
+
+    /**Determine if the value of the object is contained in the set. The
+       object values are compared, not the pointers.  So the objects in the
+       collection must correctly implement the #PObject::Compare()#
+       function. The hash table is used to locate the entry.
+
+       @return
+       TRUE if the object value is in the set.
+     */
+    BOOL operator[](
+      const T & key  /// Key to look for in the set.
+    ) const { return AbstractContains(key); }
 
     /**Get the key in the set at the ordinal index position.
     
@@ -636,15 +723,15 @@ template <class T> class PSet : public PAbstractSet
   public: \
     inline cls(BOOL initialDeleteObjects = FALSE) \
       : PAbstractSet() { AllowDeleteObjects(initialDeleteObjects); } \
-    inline virtual PObject * Clone() const \
+    virtual PObject * Clone() const \
       { return PNEW cls(0, this); } \
     inline void Include(const PObject * key) \
       { Append((PObject *)key); } \
     inline void Exclude(const PObject * key) \
       { Remove(key); } \
-    inline BOOL operator[](const K & key) \
+    inline BOOL operator[](const K & key) const \
         { return AbstractContains(key); } \
-    inline BOOL Contains(const K & key) \
+    inline BOOL Contains(const K & key) const \
         { return AbstractContains(key); } \
     virtual const K & GetKeyAt(PINDEX index) const \
       { return (const K &)AbstractGetKeyAt(index); } \
@@ -659,13 +746,16 @@ template <class T> class PSet : public PAbstractSet
   public: \
     inline cls(BOOL initialDeleteObjects = initDelObj) \
       : cls##_PTemplate() { AllowDeleteObjects(initialDeleteObjects); } \
-    inline virtual PObject * Clone() const \
+    virtual PObject * Clone() const \
       { return PNEW cls(0, this); } \
 
 
 #endif  // PHAS_TEMPLATES
 
 
+PSET(POrdinalSet, POrdinalKey);
+
+
 //////////////////////////////////////////////////////////////////////////////
 
 /**An abstract dictionary container.
@@ -712,9 +802,8 @@ class PAbstractDictionary : public PHashTable
       PObject * obj          /// New object to place into the collection.
     );
 
-    /**Insert a new object at the specified index. The index is converted to
-       a #POrdinalKey# type before being used in the #SetAt()#
-       function.
+    /**Insert a new object at the specified index. The index is as is used in
+       the #GetKeyAt()# function.
 
        @return
        #index# parameter.
@@ -724,10 +813,9 @@ class PAbstractDictionary : public PHashTable
       PObject * obj   /// New object to place into the collection.
     );
 
-    /**Remove an object at the specified index. The index is converted to
-       a #POrdinalKey# type before being used in the #GetAt()#
-       function. The returned pointer is then removed using the #SetAt()#
-       function to set that key value to NULL. If the
+    /**Remove an object at the specified index. The index is as is used in
+       the #GetKeyAt()# function. The returned pointer is then removed using
+       the #SetAt()# function to set that key value to NULL. If the
        #AllowDeleteObjects# option is set then the object is also
        deleted.
 
@@ -739,9 +827,9 @@ class PAbstractDictionary : public PHashTable
     );
 
     /**Set the object at the specified index to the new value. The index is
-       converted to a #POrdinalKey# type before being used.This will
-       overwrite the existing entry. If the AllowDeleteObjects option is set
-       then the old object is also deleted.
+       as is used in the #GetKeyAt()# function. This will overwrite the
+       existing entry. If the AllowDeleteObjects option is set then the old
+       object is also deleted.
 
        @return
        TRUE if the object was successfully added.
@@ -751,9 +839,9 @@ class PAbstractDictionary : public PHashTable
       PObject * val   /// New value to place into the collection.
     );
 
-    /**Get the object at the specified index position. The index is converted
-       to a #POrdinalKey# type before being used. If the index was not in
-       the collection then NULL is returned.
+    /**Get the object at the specified index position. The index is as is
+       used in the #GetKeyAt()# function. If the index was not in the
+       collection then NULL is returned.
 
        @return
        pointer to object at the specified index.
@@ -947,11 +1035,17 @@ template <class K, class D> class PDictionary : public PAbstractDictionary
        object is also deleted.
 
        @return
-       pointer to the object being removed, or NULL if it was deleted.
+       pointer to the object being removed, or NULL if the key was not 
+       present in the dictionary. If the dictionary is set to delete objects
+       upon removal, the value -1 is returned if the key existed prior to removal
+       rather than returning an illegal pointer
      */
     virtual D * RemoveAt(
       const K & key   /// Key for position in dictionary to get object.
-    ) { D * obj = GetAt(key); AbstractSetAt(key, NULL); return obj; }
+    ) {
+        D * obj = GetAt(key); AbstractSetAt(key, NULL); 
+        return reference->deleteObjects ? (obj ? (D *)-1 : NULL) : obj;
+      }
 
     /**Add a new object to the collection. If the objects value is already in
        the dictionary then the object is overrides the previous value. If the
@@ -1305,19 +1399,19 @@ template <class K> class POrdinalDictionary : public PAbstractDictionary
   public: \
     inline cls() \
       : PAbstractDictionary() { } \
-    inline virtual PObject * Clone() const \
+    virtual PObject * Clone() const \
       { return PNEW cls(0, this); } \
     inline PINDEX operator[](const K & key) const \
       { return (POrdinalKey &)GetRefAt(key); } \
     virtual BOOL Contains(const K & key) const \
       { return AbstractContains(key); } \
-    inline virtual POrdinalKey * GetAt(const K & key) const \
+    virtual POrdinalKey * GetAt(const K & key) const \
       { return (POrdinalKey *)AbstractGetAt(key); } \
-    inline virtual BOOL SetDataAt(PINDEX index, PINDEX ordinal) \
+    virtual BOOL SetDataAt(PINDEX index, PINDEX ordinal) \
       { return PAbstractDictionary::SetDataAt(index, PNEW POrdinalKey(ordinal)); } \
-    inline virtual BOOL SetAt(const K & key, PINDEX ordinal) \
+    virtual BOOL SetAt(const K & key, PINDEX ordinal) \
       { return AbstractSetAt(key, PNEW POrdinalKey(ordinal)); } \
-    inline virtual PINDEX RemoveAt(const K & key) \
+    virtual PINDEX RemoveAt(const K & key) \
       { PINDEX ord = *GetAt(key); AbstractSetAt(key, NULL); return ord; } \
     inline const K & GetKeyAt(PINDEX index) const \
       { return (const K &)AbstractGetKeyAt(index); } \
@@ -1340,6 +1434,4 @@ template <class K> class POrdinalDictionary : public PAbstractDictionary
 
 #endif // PHAS_TEMPLATES
 
-
-
 // End Of File ///////////////////////////////////////////////////////////////
